#ifndef GS_STRUCT_H
#define GS_STRUCT_H

#include <stdint.h>

#define GS_TOOLS_VERSION "12.09.2012 (eis_os)"

/* info about CiM file formats
 *
 * last update: 02.03.2012
 *
 * Changelog moved to changelog.txt 
 * 
 * CiM uses a chunk-based file format (similiar to microsoft's RIFF)
 * each chunk starts with a header(struct gs_chunk) that determines
 * the type and size of the following data.
 * for more infos on the individual data chunks see the comments below.
 *
 * note: every chunk always start with the chunk header - the comments below
 *       always describe the data following the header
 *
 * FILETYPE_* chunks work the same as the other chunks - just named differently
 *
 * chunks in brackets are optional
 *
 */

struct gs_chunk {
    uint32_t type;
    uint32_t size; // size of following data in bytes
} __attribute__((packed));

struct gs_vec {
    float x;
    float y;
    float z;
} __attribute__((packed));
typedef struct gs_vec gs_vec;

struct gs_quat {
    float x;
    float y;
    float z;
    float w;
} __attribute__((packed));
typedef struct gs_quat gs_quat;

struct gs_color {
    union {
        uint32_t abgr;
        struct {
            uint8_t r;
            uint8_t g;
            uint8_t b;
            uint8_t a;
        };
    };
};
typedef struct gs_color gs_color;

/*****************
 * archive files (*.gs)
 *
 *  FILETYPE_GS_ARCHIVE =:  <CHUNK_GS_ARCHIVE>
 *                          <CHUNK_EASTER_EGG>
 *
 *     CHUNK_GS_ARCHIVE =:  <CHUNK_FILE_INDEX>
 *                          <CHUNK_PACKED_DATA>
 *
 *     CHUNK_FILE_INDEX =:  UINT32 NumberOfFiles
 *			    <CHUNK_FILE_ENTRY> * NumberOfFiles
 *
 *     CHUNK_FILE_ENTRY =:  struct gs_file_entry  (size: 144 bytes)
 *
 *    CHUNK_PACKED_DATA =:  packed(zlib) data. offsets to the start
 *                          of data blocks for each files are described
 *                          in the file index
 */
#define FILETYPE_GS_ARCHIVE	0x30315347  // 'GS10'
#define CHUNK_GS_ARCHIVE	0x01000000
#define CHUNK_DUMMYDATA         0x00000001
#define CHUNK_FILE_INDEX	0x01100000
#define CHUNK_FILE_ENTRY	0x01120000
#define CHUNK_FILE_BASEPATH     0x01130000
#define CHUNK_FILE_FULLPATH     0x01140000
#define CHUNK_PACKED_DATA	0x01200000
struct gs_file_entry {
    uint32_t signature;
    uint32_t offset;	// relative to the start of CHUNK_GS_ARCHIVE data
    uint32_t size_unpacked;
    uint32_t size_packed;
    char name[128];
} __attribute__((packed));



/****************************
 *  object files (*.object)
 *
 *   FILETYPE_GS_OBJECT =:  struct gs_object   (size: 60 bytes)
 *                          <CHUNK_OBJECT_DATA>
 *                         (<CHUNK_OBJECT_DATA2>)  ???
 *                         (<CHUNK_OBJ_PATH_INFO>)
 *
 *
 *    CHUNK_OBJECT_DATA =:  <CHUNK_VERTEX_DATA>
 *                          <CHUNK_INDEX_DATA>
 *                          <CHUNK_MATERIAL_DATA>
 *
 *   CHUNK_OBJECT_DATA2 =:  same structure as CHUNK_OBJECT_DATA
 *
 *  CHUNK_OBJ_PATH_INFO =:  struct gs_path
 *
 *
 *    CHUNK_VERTEX_DATA =:  UINT32 NumberOfVertices
 *                          struct gs_vertex * NumberOfVertices
 *                         (<CHUNK_VERTEX_EXTRA>) ??
 *                         (<CHUNK_VERTEX_EXTRA2>) ??
 *
 *     CHUNK_INDEX_DATA =:  UINT32 NumberOfIndices
 *                          UINT32 idx * NumberOfIndices
 *
 *  CHUNK_MATERIAL_DATA =:  UINT32 NumberOfMaterials
 *			    <CHUNK_MATERIAL_DEF> * NumberOfMaterials
 *
 *
 *  CHUNK_VERTEX_EXTRA  =: TODO
 *  CHUNK_VERTEX_EXTRA2 =: TODO
 *   CHUNK_MATERIAL_DEF =: see material files
 *
 */

#define FILETYPE_GS_OBJECT	0x02000000
#define CHUNK_OBJECT_DATA	0x02010000
#define CHUNK_OBJECT_DATA2      0x02040000   //OBJ_MORE_VERTEX
#define CHUNK_OBJ_PATH_INFO	0x02600000
#define CHUNK_VERTEX_DATA	0x02011000
#define CHUNK_INDEX_DATA	0x02012000
#define CHUNK_MATERIAL_DATA	0x02013000
#define CHUNK_MATERIAL_DEF	0x02013100
#define CHUNK_VERTEX_EXTRA2     0x02011100      //GS_CHUNK_WEIGHTS
#define CHUNK_VERTEX_EXTRA	0x02011200      //GS_CHUNK_UVCHANNEL

struct gs_object {
    char name[32];
    gs_vec pos;
    gs_quat quat;
} __attribute__((packed));

struct gs_path {        // describes a path node
    uint32_t flags;     // flags by gebbiz, no additional info known
    char name[32];
    gs_vec pos;
    gs_quat quat;       // quaternion rotation
} __attribute__((packed));

struct gs_vertex {
    gs_vec pos;
    gs_vec nor;
    uint32_t color; // vertex color ?? seems unused
    float uv[2];
} __attribute__((packed));


/*******************
 * material files (*.material)
 *
 * basically just a CHUNK_MATERIAL_DEF in its own file
 * difference is that only here it contains texture infos
 *
 * FILETYPE_GS_MATERIAL ==  CHUNK_MATERIAL_DEF
 *
 *   CHUNK_MATERIAL_DEF =:  struct gs_material
 *			   (<CHUNK_MAT_TEXTURE>) // only in material file
 *			   (<CHUNK_MAT_TEXTURE2>)     ""    (bump map?)
 *			   (<CHUNK_MAT_TEXTURE3>)     ""    (spec map?)
 *
 *    CHUNK_MAT_TEXTURE =:  struct gs_tex
 *			   (<CHUNK_TEX_ORIG_FILE>)   // leftover from converting
 *			   (<CHUNK_TEX_ORIG_FILE2>)  // process ??
 *   CHUNK_MAT_TEXTURE2 =: same struct as CHUNK_MAT_TEXTURE
 *   CHUNK_MAT_TEXTURE3 =: same struct as CHUNK_MAT_TEXTURE
 *
 *  CHUNK_TEX_ORIG_FILE =: char[]
 * CHUNK_TEX_ORIG_FILE2 =: char[]
 *
 */

#define FILETYPE_GS_MATERIAL	CHUNK_MATERIAL_DEF
#define CHUNK_MAT_TEXTURE	0x02013110      // Gebbiz: DIFFUSEMAP
#define CHUNK_MAT_TEXTURE2      0x02013120      // Gebbiz: NORMALMAP
#define CHUNK_MAT_NORMALMAP     0x02013120      // Gebbiz: NORMALMAP
#define CHUNK_MAT_TEXTURE3	0x02013130      // Gebbiz: SPECULARMAP
#define CHUNK_MAT_SPECULARMAP   0x02013130
#define CHUNK_MAT_ILLUMMAP      0x02013140
#define CHUNK_MAT_AOMAP         0x02013150      // ambient occlusion?
#define CHUNK_TEX_ORIG_FILE	0x02013111
#define CHUNK_TEX_ORIG_FILE2	0x02013112


struct gs_material {
    char name[32];
//    uint32_t unk[6]; //probably describes material colors(seems always the same)
    uint8_t type;
    uint8_t flags;
    uint8_t twoSided;   // bool
    uint8_t opacity;
    uint8_t specularLevel;
    uint8_t glossiness;
    uint8_t brightness;
    uint8_t reflection;
    gs_color ambient;
    gs_color diffuse;
    gs_color specular;
    gs_color illumination;    
    uint32_t vert_start, vert_num; // describes to which vertices/idx in
    uint32_t idx_start, idx_num;   // VERTEX/INDEX_DATA this material applies
} __attribute__((packed));

struct gs_tex {
    /* uint32_t a, b; // ??
     * 
     */
    uint8_t type;
    uint8_t amount;
    uint8_t alphaChannel;
    uint8_t alphaBits;
    uint8_t tileU;                // bool
    uint8_t tileV;                // bool
    uint8_t mirrorU;              // bool
    uint8_t mirrorV;              // bool
    char file[32];
} __attribute__((packed));



/************************
 * scene files (*.scene)
 *
 * - not really examined yet
 *
 *    FILETYPE_GS_SCENE =:  <CHUNK_SCENE_ORIG_FILE>
 *                          <CHUNK_SCENE_DESCR>
 *
 *CHUNK_SCENE_ORIG_FILE =:  char[]  // name of original FBX file
 *
 *    CHUNK_SCENE_DESCR =:  struct gs_scene_descr
 *
 */

#define FILETYPE_GS_SCENE	0x02100000
#define CHUNK_SCENE_ORIG_FILE	0x02100001
#define CHUNK_SCENE_DESCR	0x02080000 //??

struct gs_scene_descr {
    uint32_t unk;
    char     name[32];
    char     file[64];
    float    ofX, ofY, ofZ; // just a guess
    uint32_t u1, u2, u3, u4; // no clue
} __attribute__((packed));


/*************
 * skeleton files (*.skeleton)
 *
 * not really sure what exactly is described in here but it seems to
 * belong together with the PATH_INFO stuff in the object files
 *
 * FILETYPE_GS_SKELETON =:  char[32] name
 *                          UINT32 NumberOfBones
 *                          <CHUNK_BONE> * NumberOfBones
 *
 *           CHUNK_BONE =:  struct gs_bone
 *                         (<CHUNK_BONE_OBJ_REF>)
 *
 *   CHUNK_BONE_OBJ_REF =:  char[]
 *
 */

#define FILETYPE_GS_SKELETON	0x02200000
#define CHUNK_BONE		0x02201000
#define CHUNK_BONE_OBJ_REF	0x02201010

struct gs_bone {
    char name[32];
    int32_t n;   //??
    float u[10]; // ??
} __attribute__((packed));


/* other files */
#define FILETYPE_GS_ANIMATION	0x02300000
#define FILETYPE_GS_PROPERTIES	0x02400000
#define FILETYPE_GS_GRID	0x07000000

// the map chunk has map x*y*4 bytes
#define CHUNK_GS_GRID_MAP	0x07010000
#define CHUNK_GS_GRID_OBJECTS	0x07020000
#define CHUNK_GS_GRID_LISTOBJECTS         0x07020100
#define CHUNK_GS_GRID_OBJ_INSTANCE        0x07020210
#define CHUNK_GS_GRID_OBJ_INSTANCE_NAMES  0x07020310

//#define CHUNK_GS_GRID_UNKNOWN3	   0x07030000

#define CHUNK_GS_GRID_UNKNOWN4	   0x07040000
#define CHUNK_GS_GRID_LISTVEHICLES 0x07040100
#define CHUNK_GS_GRID_UNKNOWN4_2   0x07040200
        
#define CHUNK_GS_GRID_UNKNOWN5	   0x07050000
#define CHUNK_GS_GRID_LISTCITIZEN  0x07050100
#define CHUNK_GS_GRID_UNKNOWN5_2   0x07050200

#define CHUNK_GS_GRID_UNKNOWN6	0x07060000
#define CHUNK_GS_GRID_LISTPLAYERVEHICLES  0x07060100

#define CHUNK_GS_GRID_UNKNOWN7	      0x07070000
#define CHUNK_GS_GRID_UNKNOWN7_1      0x07070100

// Types between 0x0702 ... 0x0707 found...

struct gs_grid {
    uint32_t u1, u2;   // could be X, Y map size?
    uint32_t u3, u4;
    uint32_t u5, u6;
} __attribute__((packed));

struct gs_grid_map {
    int16_t height;    // seems to wrap around, so assumeing signed
    uint8_t txid;
    uint8_t flags;     // 0x20 = water;
} __attribute__((packed));


struct gs_grid_objects_header {
    uint32_t numobjectids;
    uint32_t numinstances;
} __attribute__((packed));

struct gs_grid_objects_instance {
    uint16_t refnum;    // 0x3FFF = the reference number in CHUNK_GS_GRID_LISTOBJECTS
                        // high bits rotation;
    int16_t posX;       // in slots
    int16_t posY;       // height above ground
    int16_t posZ;
    uint32_t u1;
    uint32_t u2;
} __attribute__((packed));

struct gs_grid_objects_instance3 {
    uint32_t u1;
} __attribute__((packed));

struct gs_grid_commonlist_header {
     uint32_t numlist;
     uint32_t u2;
} __attribute__((packed));

struct gs_grid_unknown6_header {
     uint32_t numplayervehiclenames;
     uint32_t u2;
     uint32_t u3;
     uint32_t u4;
} __attribute__((packed));

/*************
 * timeline files (*.timeline)
 *
 *
 * FILETYPE_GS_TIMELINE =:  gs_timeline
 *                          <CHUNK_GS_TIMELINE_ENTRY> * ?
 *
 *           CHUNK_GS_TIMELINE_ENTRY =:  struct gs_timeline_entry
 *                                      (<CHUNK_GS_TIMELINE_EUNK1>)
 *
 *
 */

#define FILETYPE_GS_TIMELINE	0x07080000
#define CHUNK_GS_TIMELINE_ENTRY	0x07081000
#define CHUNK_GS_TIMELINE_EUNK1	0x07081010
#define CHUNK_GS_TIMELINE_EUNK2	0x07081020
#define CHUNK_GS_TIMELINE_EUNK4	0x07081040
#define CHUNK_GS_TIMELINE_EUNK8	0x07081080
#define CHUNK_GS_TIMELINE_EUNK7	0x07081071

struct gs_timeline {
     uint32_t days, u2;
} __attribute__((packed));

// CHUNK_GS_TIMELINE_ENTRY:
struct gs_timeline_entry {
     uint32_t u1, u2;
} __attribute__((packed));

struct gs_timeline_eunk1 {
     char objectname[32];
} __attribute__((packed));

struct gs_timeline_eunk2 {
     char objectname[32];
     uint32_t u1, u2, u3;
} __attribute__((packed));

struct gs_timeline_eunk4 {
     char objectname[32];
     uint32_t u1, u2, u3;
} __attribute__((packed));

struct gs_timeline_eunk8 {
     char objectname[32];
     uint32_t u1, u2, u3, u4;
} __attribute__((packed));

struct gs_object_properties {
  char name[32];
  char className[32];
  uint8_t sortAlpa;
  uint8_t defaultDrawDistance;
  uint16_t reserved;
  float minDrawDistance;
  float maxDrawDistance;
} __attribute__((packed));


struct gs_states_entry {
  char name[32];
  uint32_t u1, u2, u3, u4;
} __attribute__((packed));


#define FILETYPE_GS_STATES           0x05000000
#define CHUNK_GS_STATES_NAMES        0x05010000
#define CHUNK_GS_STATES_NAME         0x05010100
#define CHUNK_GS_STATES_SCRIPTS      0x05020000
#define CHUNK_GS_STATES_ENTRY        0x05020100
#define CHUNK_GS_STATES_ENTRYSCRIPT  0x05020101
#define CHUNK_GS_STATES_ENTRYSCRIPT2  0x05020102
#endif
