/*
 * gs compress - packing files in gs format
 *
 * by eis_os (cim bytetransfer.de)
 * 
 * last update: 02.03.2012
 *
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <zlib.h>
#include <assert.h>

#include "gs.h"
#include "gs_filefunc.h"

#define TOOLNAME "gs_compress " GS_TOOLS_VERSION 

// options
static int force_overwrite = 0;

void usage(char *name) {
    printf("Usage: %s <options> directory <file>\n"
    " -f force overwrite\n"
    " -b basedirectory\n"
    " -p fullpath\n"
    " --exclude pattern\n", name);
}

struct excludelist {
    char* pattern;
    struct excludelist *prev;
};


struct filelist {
    char *filename;
    char *filepath;
    char *path;
    uint32_t fsize;
    uint32_t fsize_compressed;
    uint32_t headerbytes;
    
    /* for reading on filesytem */
    char *fullfilepath;
   
    /* location in output stream */
    uint32_t filelocation;
    
    struct filelist *next;
    struct filelist *prev;
};


int matcher(char* p, char* subject) {
    //print(1, "Matching %s %s\n", p, subject);
    for ( ; *p; subject++, p++) {
        if (*subject == '\0' && *p != '*')
	    return 0;
        switch(*p) {
            case '*':
                while ( (*++p == '*' || *p == '?') ) {
                    if (*p == '*') continue;
                    if (*p == '?' && *subject == '\0') return 0;
                };
                if (*p == '\0') return 1;
                while(*subject != '\0') {
                        // *p is at a normal char
                        if (*subject == *p) {
                                if (matcher(p, subject) == 1) return 1;
                        }
                        subject++;
                }
                return 0;
            case '?':
                if (subject == '\0') return 0;
                continue;
            default:
                if (*p != *subject) return 0;
        }
    }
    if (*subject == '\0') return 1;
    return 0;
}

struct filelist* getfilelist(char *path, char* subpath, struct filelist *flistprev, struct excludelist* exclude) {
    char apath[4096];
    char asubpath[4096];
    DIR *dir;
    struct dirent *dirEntry;
    struct stat attribut;
    int skipfile = 0;
    
    int subpathlen = 0;
    if (subpath) {
         subpathlen = strlen(subpath);
    }
  
    //printf( "Scanning %s\n", path);
    if( ( dir = opendir( path ) ) != NULL ) {
        while( ( dirEntry = readdir( dir ) ) != NULL ) {
            if(strcmp( (char*)dirEntry->d_name, "." ) == 0 || strcmp( (char*)dirEntry->d_name, ".." ) == 0 ) continue;
            
            strcpy(apath, path);
            strcat(apath, "\\");
            strcat(apath, dirEntry->d_name);

            if( stat( apath, &attribut ) == -1 ) {
                die("stat(%s) failed\n", apath );
            }
            
            if( ( attribut.st_mode & S_IFDIR ) ) {
                asubpath[0] = '\0';
                if (subpath) {
                        strcpy(asubpath, subpath);
                        strcat(asubpath, "/");
                }
                strcat(asubpath, dirEntry->d_name);
                //printf( "Scanning dir %s\n", apath);
               
                flistprev = getfilelist(apath, asubpath, flistprev, exclude);
            } else {
                struct filelist *flist = calloc(1, sizeof( struct filelist));        
                flist->filename = strdup(dirEntry->d_name);
                if (subpath) {
                        flist->path = strdup(subpath);
                } else {
                        flist->path = strdup("");
                }
                flist->filepath = calloc(subpathlen+1+1+strlen(flist->filename)+1, 1);
                if (subpath) {
                    strcat(flist->filepath, subpath);
                    strcat(flist->filepath, "/");
                }
                strcat(flist->filepath, dirEntry->d_name);
                skipfile = 0;
                if (exclude) {
                    struct excludelist *exlist;
                    exlist = exclude;
                    do {
                        if (matcher(exlist->pattern, flist->filepath) == 1) {
                             skipfile = 1;
                             break;
                        }
                    } while(exlist = exlist->prev);
                }
                if (skipfile) {
                    continue;
                }
                
                flist->fullfilepath = strdup(apath);
                flist->prev = flistprev;
                
                if (flistprev) {
                    flistprev->next = flist;
                }
                flistprev = flist;
               
                //printf( "%s in %s\n", dirEntry->d_name, flist->filepath);
                //printf("fileentry: %02x\n", flist);
            }
        }
        closedir(dir);
        return flistprev;
    }
    die("Can't open directory %s", path);
}


void creategsarchive(char* filename, struct filelist *flist, char* basepath, char* fullpath, char* footerinfo) {
    FILE* f;
    void *ch_filetype_gs_archive;
    void *ch_gs_archive;
    void *ch_gs_fileindex;
    void *ch_gs_packdata;
    uint32_t gsarchive_offset;
    uint32_t gsfileindex_start;
    uint32_t gsarchive_end;
    struct gs_chunk ch;
    
    uint32_t num_files = 0;
    struct gs_file_entry fe;
    struct filelist *flistptr = flist;
    while(flistptr != NULL) {
        num_files++;
        flistptr = flistptr->next;
    }
    flistptr = flist;
    
    print(0, "Writing gs archive file: %s\n", filename);
    
    if (basepath && fullpath) {
        die("Writeing basepath(%s) and fullpath(%s) is not possible\n", basepath, fullpath);
    }
    if(!force_overwrite && file_exists(filename))
        die("'%s' already exists. Remove or use -f to force overwrite\n", filename);
    
    f = fopen(filename, "wb");
    
    ch_filetype_gs_archive = fwrite_chunk_start(f, FILETYPE_GS_ARCHIVE);
        ch_gs_archive = fwrite_chunk_start(f, CHUNK_GS_ARCHIVE);
            gsarchive_offset = ftell(f);
            ch_gs_fileindex = fwrite_chunk_start(f, CHUNK_FILE_INDEX);
                fwrite(&num_files, sizeof(num_files), 1, f);
                
                if (basepath) {
                    print(1, "Writeing basepath(%s) entry\n", basepath);
                    void* ch_gs_basepath = fwrite_chunk_start(f, CHUNK_FILE_BASEPATH);
                    char* basepath128 = malloc(128);
                    strncpy(basepath128, basepath, 128);
                    fwrite(basepath128, 128, 1, f);
                    free(basepath128);
                    fwrite_chunk_end(ch_gs_basepath);
                }
                
                if (fullpath) {
                    print(1, "Writeing fullpath(%s) entry\n", basepath);
                    void* ch_gs_fullpath = fwrite_chunk_start(f, CHUNK_FILE_FULLPATH);
                    char* fullpath128 = malloc(128);
                    strncpy(fullpath128, basepath, 128);
                    fwrite(fullpath128, 128, 1, f);
                    free(fullpath128);
                    fwrite_chunk_end(ch_gs_fullpath);
                }
                gsfileindex_start = ftell(f);
                print(1, "Alloc room for file index table\n");
                fseek(f, (num_files*(sizeof(fe)+sizeof(ch))), SEEK_CUR);
            fwrite_chunk_end(ch_gs_fileindex);
            ch_gs_packdata = fwrite_chunk_start(f, CHUNK_PACKED_DATA);

                print(1, "Write compressing file stream:\n");
                flistptr = flist;
                while(flistptr != NULL) {
                    print(2, "%s   ", flistptr->filepath);
                    compressfile(flistptr, f);
                    print(3, "size: %u compressedsize: %u\n", flistptr->fsize, flistptr->fsize_compressed);
                    flistptr = flistptr->next;
                }
            fwrite_chunk_end(ch_gs_packdata);
            gsarchive_end = ftell(f);
            
            fseek(f, gsfileindex_start, SEEK_SET);
            print(1, "writing index entries for files:\n");
            flistptr = flist;
            while(flistptr != NULL) {
                print(2, "%s\n", flistptr->filepath);
                memset(&fe, 0, sizeof(fe));
                strncpy(fe.name, flistptr->filepath, 128);
                fe.signature = flistptr->headerbytes;
                fe.size_packed = flistptr->fsize_compressed;
                fe.size_unpacked = flistptr->fsize;
                fe.offset = flistptr->filelocation - gsarchive_offset;

                ch.type = CHUNK_FILE_ENTRY;
                ch.size = sizeof(fe);
                fwrite(&ch, sizeof(ch), 1, f);
                fwrite(&fe, sizeof(fe), 1, f);
                flistptr = flistptr->next;
            }
            fseek(f, gsarchive_end, SEEK_SET);
        fwrite_chunk_end(ch_gs_archive);
    // easter egg?
    if (footerinfo) {
        void* ch_gs_info = fwrite_chunk_start(f, CHUNK_DUMMYDATA);
        fwrite(footerinfo,1,strlen(footerinfo), f);
        fwrite_chunk_end(ch_gs_info);
    }
   
    fwrite_chunk_end(ch_filetype_gs_archive);
    fclose(f);
}

#define CHUNK 16384
int compressfile(struct filelist *fentry, FILE* dest) {
    FILE* source;
    int ret, flush;
    unsigned have;
    z_stream strm;
    unsigned char in[CHUNK];
    unsigned char out[CHUNK];

    int level = -1;
    
    fentry->filelocation = ftell(dest);
    
    source = fopen(fentry->fullfilepath, "rb");
    fseek(source, 0, SEEK_SET);
    fread(&fentry->headerbytes, sizeof(fentry->headerbytes), 1, source);
    fseek(source, 0, SEEK_END);
    fentry->fsize = ftell(source);
    fseek(source, 0, SEEK_SET);
            
    /* allocate deflate state */
    strm.zalloc = Z_NULL;
    strm.zfree = Z_NULL;
    strm.opaque = Z_NULL;
        
    ret = deflateInit(&strm, level);
    if (ret != Z_OK)
        return ret;

    /* compress until end of file */
    do {
        strm.avail_in = fread(in, 1, CHUNK, source);
        if (ferror(source)) {
            (void)deflateEnd(&strm);
            return Z_ERRNO;
        }
        flush = feof(source) ? Z_FINISH : Z_NO_FLUSH;
        strm.next_in = in;

        /* run deflate() on input until output buffer not full, finish
           compression if all of source has been read in */
        do {
            strm.avail_out = CHUNK;
            strm.next_out = out;
            ret = deflate(&strm, flush);    /* no bad return value */
            assert(ret != Z_STREAM_ERROR);  /* state not clobbered */
            have = CHUNK - strm.avail_out;
            if (fwrite(out, 1, have, dest) != have || ferror(dest)) {
                (void)deflateEnd(&strm);
                return Z_ERRNO;
            }
        } while (strm.avail_out == 0);
        assert(strm.avail_in == 0);     /* all input will be used */

        /* done when last data in file processed */
    } while (flush != Z_FINISH);
    assert(ret == Z_STREAM_END);        /* stream will be complete */

    /* clean up and return */
    (void)deflateEnd(&strm);
    fentry->fsize_compressed = ftell(dest) - fentry->filelocation;
    
    fclose(source);
    return 0;
}

int main(int argc, char **argv)
{
    int i;
    char *filename = NULL, *directory = NULL;
    char* basefolder = NULL;
    char* fullpath = NULL;
    char* footerinfo = NULL;
    
    struct filelist *flist = NULL;
    struct excludelist *elist = NULL;
    struct stat attribut;
    
    printf(TOOLNAME "\n");
    if(argc < 2) {
	usage(argv[0]);
	return -1;
    }
    
    for(i=1; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
            if (arg[1] == 'f') {
                force_overwrite = 1;
            } else if (arg[1] == 'b') {
                i++;
                if (i<argc) {
                        basefolder = strdup(argv[i]);
                } else {
                        usage(argv[0]);
                        return -1;
                }
            } else if (arg[1] == 'p') {
                i++;
                if (i<argc) {
                        fullpath = strdup(argv[i]);
                } else {
                        usage(argv[0]);
                        return -1;
                }
            } else if (arg[1] == 'i') {
                i++;
                if (i<argc) {
                        footerinfo = strdup(argv[i]);
                } else {
                        usage(argv[0]);
                        return -1;
                }
            } else if (arg[1] == '-') {
                if (strcmp(arg,"--exclude") == 0) {
                        i++;
                        if (i<argc) {
                            struct excludelist *e;
                            e = malloc(sizeof(struct excludelist));
                            e->pattern = strdup(argv[i]);
                            e->prev = elist;
                            elist = e;
                            print(1, "Excluding: %s\n", e->pattern);
                        } else {
                            usage(argv[0]);
                            return -1;
                        }
                }
            } else {
                printf("Unknown option %s \n", argv[i]);
                return -1;
            }
	} else {
            if(directory == NULL) {
                if( stat(argv[i], &attribut ) == -1 ) {
                    die("Can't determine type of %s, stat failed\n", argv[i]);
                }
                if((S_ISDIR(attribut.st_mode)) ) {
                    directory = strdup(argv[i]);
                } else {
                    die("%s doesn't seem to be a directory\n", argv[i]);
                }
            } else {
                filename = strdup(argv[i]);
            }
        }
    }
    if (!directory) {
        usage(argv[0]);
        return -1;
    }    
    if (!filename) {
        filename = calloc(strlen(directory)+4, 1);
        strcpy(filename, directory);
        strcat(filename, ".gs");
    }
        
    struct filelist *oldflist;
    flist = getfilelist(directory, NULL, NULL, elist);

    // rewind filelist, sanity checks
    while(flist != NULL) {
        //printf("(%02x) %s   flist->next = (%02x)\n", flist, flist->fullfilepath, flist->next);
        oldflist = flist;
        if (flist->prev) {
            flist = flist->prev;
            if (flist->next != oldflist) {
                die("!!error: flist->next(%02x) != oldflist(%02x)\n", flist, oldflist);
            } 
        } else {
            break;
        }   
    }
    creategsarchive(filename, flist, basefolder, fullpath, footerinfo); 
    return 0;
}