/*
 * gs dump utility - trying to decipher the Cities In Motion game data
 *
 * by immi ( immi at gmx . org )
 * by eis_os (cim bytetransfer.de)
 * 
 * last update: 20.03.2012
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <errno.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "gs.h"
#include "gs_filefunc.h"

#define TOOLNAME "gs_dump " GS_TOOLS_VERSION 

// options
int verbose = 0;
int fileoffsets = 0;

static void chunk_pr(struct gs_chunk *ch, const char *name, int level)
{
    print(level, "%s(%#08x) (%u bytes)\n", name ? name : "UNKNOWN_CHUNK",
	  ch->type, ch->size);
}

static void chunk_pr_ext(struct gs_chunk *ch, const char *name, int level, int fd)
{
    if (fileoffsets) {
        print(level, "@ %#0.8x %s(%#08x) (%u bytes) \n", tell(fd)-8, name ? name : "UNKNOWN_CHUNK", ch->type, ch->size);
    } else {
        print(level, "%s(%#08x) (%u bytes)\n", name ? name : "UNKNOWN_CHUNK", ch->type, ch->size);
    }
}


static void dump_gs_archive(int fd, unsigned int size)
{
    uint32_t num_files;
    struct gs_chunk ch, ch2, ch3;
    struct gs_file_entry fe;
    uint32_t gsarchive_offset;

    while(size) {
	gs_read_chunk(fd, &ch, &size);

	if(ch.type != CHUNK_GS_ARCHIVE) {
	    chunk_pr_ext(&ch,0, 1, fd); // unknown
	    lseek(fd, ch.size, SEEK_CUR);
 
	    continue;
	}
        gsarchive_offset = tell(fd);
	chunk_pr_ext(&ch, "CHUNK_GS_ARCHIVE", 1, fd);
	while(ch.size) {
	    gs_read_chunk(fd, &ch2, &ch.size);

	    switch(ch2.type) {
	    case CHUNK_FILE_INDEX:
		chunk_pr_ext(&ch2, "CHUNK_FILE_INDEX", 2, fd);
		gs_read(fd, &num_files, sizeof(num_files), &ch2.size);
		print(2, "NumberOfFiles: %u\n", num_files);

		while(ch2.size) {

		    gs_read_chunk(fd, &ch3, &ch2.size);

                    if(ch3.type == CHUNK_FILE_ENTRY) {
                        chunk_pr_ext(&ch3, "CHUNK_FILE_ENTRY", 3, fd);
                        gs_read(fd, &fe, sizeof(fe), &ch3.size);

                        print(3, "%6u (%6u)  %.128s  -> %#08x realofs = %#08x\n", fe.size_packed,
                              fe.size_unpacked, fe.name, fe.offset, gsarchive_offset + fe.offset);
                    } else if (ch3.type == CHUNK_FILE_BASEPATH) {
                        chunk_pr_ext(&ch3, "CHUNK_FILE_BASEPATH", 3, fd);
                        char *buf = calloc(ch3.size+1, 1);
                        gs_read(fd, buf, ch3.size, &ch3.size);
                        print(3, "%.128s\n", buf);
                        free(buf);
                    } else if (ch3.type == CHUNK_FILE_FULLPATH) {
                        chunk_pr_ext(&ch3, "CHUNK_FILE_FULLPATH", 3, fd);
                        char *buf = calloc(ch3.size+1, 1);
                        gs_read(fd, buf, ch3.size, &ch3.size);
                        print(3, "%.128s\n", buf);
                        free(buf);                        
                    } else {
                        chunk_pr_ext(&ch3, 0, 3, fd);
			lseek(fd, ch3.size, SEEK_CUR);
                    }
		}
		break;
	    case CHUNK_PACKED_DATA:
		chunk_pr_ext(&ch2, "CHUNK_PACKED_DATA", 2, fd);
		lseek(fd, ch2.size, SEEK_CUR);
		break;
	    default:
		chunk_pr_ext(&ch2, 0, 2, fd);
		lseek(fd, ch2.size, SEEK_CUR);
	    }
	}
    }
}

static void dump_vertex(struct gs_vertex *v)
{
    printf(" pos[%.6g, %.6g, %.6g] nor[%.6g, %.6g, %.6g]"
	   " uv[%.6g, %.6g] color:%#08x\n", v->pos.x, v->pos.y,v->pos.z,
	   v->nor.x, v->nor.y, v->nor.z, v->uv[0], v->uv[1], v->color);
}

static void dump_vertices(int fd, unsigned int num, unsigned int *rem)
{
    unsigned int i;
    struct gs_vertex v;

    if(verbose) {
	for(i = 0; i < num; i++) {
	    struct gs_vertex v;
	    gs_read(fd, &v, sizeof(v), rem);
	    dump_vertex(&v);
	}
    } else {
	lseek(fd, num * sizeof(v), SEEK_CUR);
	*rem -= num * sizeof(v);
	print(3, "<...use -l to see all vertex data>\n");
    }
}

static void dump_vertex_data(int fd, struct gs_chunk *ch, unsigned int level)
{
    uint32_t num;

    chunk_pr(ch, "CHUNK_VERTEX_DATA", level++);

    gs_read(fd, &num, sizeof(num), &ch->size);
    print(level, "NumberOfVertices: %u\n", num);

    dump_vertices(fd, num, &ch->size);

    while(ch->size) {
	struct gs_chunk extra;
	gs_read_chunk(fd, &extra, &ch->size);
	if(extra.type == CHUNK_VERTEX_EXTRA)
	    chunk_pr(&extra, "CHUNK_VERTEX_EXTRA", level);
	else if(extra.type == CHUNK_VERTEX_EXTRA2)
	    chunk_pr(&extra, "CHUNK_VERTEX_EXTRA2", level);
	else
	    chunk_pr(&extra, 0, 3);

	lseek(fd, extra.size, SEEK_CUR);
    }
}

static void dump_indices(int fd, unsigned int num, unsigned int *rem)
{
    unsigned int i;
    uint32_t idx;

    if(verbose) {
        for(i = 0; i < num; i++) {
	    gs_read(fd, &idx, sizeof(idx), rem);
	    printf(" idx: %u\n", idx);
	}
    } else {
	*rem -= num * sizeof(idx);
	lseek(fd, num * sizeof(idx), SEEK_CUR);
	print(3, "<... use -l to see all index data\n");
    }

}

static void dump_tex(int fd, unsigned int *len, unsigned int level)
{
    struct gs_tex tex;

    gs_read(fd, &tex, sizeof(tex), len);

    print(level, "texture: '%.32s'\n", tex.file);

    print(level, "type:          0x%02x       amount: 0x%02x\n", tex.type, tex.amount);
    print(level, "alphaChannel:  0x%02x    alphaBits: 0x%02x\n", tex.alphaChannel, tex.alphaBits);
    print(level, "tileU(b):      0x%02x     tileV(b): 0x%02x\n", tex.tileU, tex.tileV);
    print(level, "mirrorU(b):    0x%02x   mirrorV(b): 0x%02x\n", tex.mirrorU, tex.mirrorV);
    


    while(*len) {
	struct gs_chunk ch;
	gs_read_chunk(fd, &ch, len);

	if(ch.type == CHUNK_TEX_ORIG_FILE) {
	    chunk_pr(&ch, "CHUNK_TEX_ORIG_FILE", level);
	    char *str = malloc(ch.size);
	    gs_read(fd, str, ch.size, &ch.size);
	    print(level+1, "file: %s\n", str);
	    free(str);
	} else if(ch.type == CHUNK_TEX_ORIG_FILE2) {
	    chunk_pr(&ch, "CHUNK_TEX_ORIG_FILE2", level);
	    char *str = malloc(ch.size);
	    gs_read(fd, str, ch.size, &ch.size);
	    print(level+1, "file: %s\n", str);
	    free(str);
	} else {
	    chunk_pr(&ch, 0, level);
	    lseek(fd, ch.size, SEEK_CUR);
	}
    }
}


static void dump_material(int fd, struct gs_chunk *ch, unsigned int level)
{
    struct gs_material mat;
    unsigned int i;
    chunk_pr(ch, "CHUNK_MATERIAL_DEF", level++);

    gs_read(fd, &mat, sizeof(mat), &ch->size);
    print(level, "name: %.32s\n", mat.name);
        
    print(level,"type:          0x%02x        flags: 0x%02x\n", mat.type, mat.flags);
    print(level,"twoSided:      0x%02x      opacity: 0x%02x\n", mat.twoSided, mat.opacity);
    print(level,"specularLevel: 0x%02x   glossiness: 0x%02x\n", mat.specularLevel, mat.glossiness);
    print(level,"brightness:    0x%02x   reflection: 0x%02x\n", mat.brightness, mat.reflection);
    
    print(level,"ambient:       0x%08x\n", mat.ambient);
    print(level,"diffuse:       0x%08x\n", mat.diffuse);
    print(level,"specular:      0x%08x\n", mat.specular);
    print(level,"illumination:  0x%08x\n", mat.illumination);

    
    print(level, "used for vertices: [%u, %u]\n",
	  mat.vert_start, mat.vert_start + mat.vert_num);
    print(level, "used for indices:  [%u, %u]\n",
	  mat.idx_start, mat.idx_start + mat.idx_num);

    while(ch->size) {
	struct gs_chunk tex;
	gs_read_chunk(fd, &tex, &ch->size);

	if(tex.type == CHUNK_MAT_TEXTURE) {
	    chunk_pr(&tex, "CHUNK_MAT_TEXTURE (DIFFUSEMAP)", level);
	    dump_tex(fd, &tex.size, level + 1);
	} else if(tex.type == CHUNK_MAT_TEXTURE2) {
	    chunk_pr(&tex, "CHUNK_MAT_TEXTURE2 (NORMALMAP)", level);
	    dump_tex(fd, &tex.size, level + 1);
	} else if(tex.type == CHUNK_MAT_TEXTURE3) {
	    chunk_pr(&tex, "CHUNK_MAT_TEXTURE3 (SPECULARMAP)", level);
	    dump_tex(fd, &tex.size, level + 1);
        } else if(tex.type == CHUNK_MAT_ILLUMMAP) {
	    chunk_pr(&tex, "CHUNK_MAT_ILLUMMAP", level);
	    dump_tex(fd, &tex.size, level + 1);
        } else if(tex.type == CHUNK_MAT_AOMAP) {
	    chunk_pr(&tex, "CHUNK_MAT_AOMAP", level);
	    dump_tex(fd, &tex.size, level + 1);
        } else
	    chunk_pr(&tex, 0, level);

	lseek(fd, tex.size, SEEK_CUR);
    }
}

static void dump_path_info(int fd, struct gs_chunk *ch, unsigned int level)
{
    struct gs_path path;
    unsigned int i;
    chunk_pr(ch, "CHUNK_OBJ_PATH_INFO", level++);
    gs_read(fd, &path, sizeof(path), &ch->size);

    print(level, "name: '%.32s' flags: %#08x\n", path.name, path.flags);
  /*  for(i = 0; i < 7; i++) {
	if(i%4 == 0) printf("\n%*sunk:",level*2,"");
	printf(" %.4g", path.unk2[i]);
    } */
    print(level, "pos:  [x: %.6g, y: %.6g, z: %.6g]\n",
        path.pos.x, path.pos.y, path.pos.z);
    print(level, "quat: [x: %.6g, y: %.6g, z: %.6g, w: %.6g]\n", 
	path.quat.x, path.quat.y, path.quat.z, path.quat.w);
    
    //printf("\n");
    if(ch->size) {
	print(level, "THERE IS MORE DATA HERE\n");
	lseek(fd, ch->size, SEEK_CUR);
    }
}


static void dump_obj_data(int fd, unsigned int *size, unsigned int level)
{
    uint32_t num;

    while(*size) {
	struct gs_chunk data;
	gs_read_chunk(fd, &data, size);

	if(data.type == CHUNK_VERTEX_DATA) {
	    dump_vertex_data(fd, &data, level);
	} else if( data.type == CHUNK_INDEX_DATA) {
	    chunk_pr(&data, "CHUNK_INDEX_DATA", level);
	    gs_read(fd, &num, sizeof(num), &data.size);
	    print(level+1, "NumberOfIndices: %u\n", num);

	    dump_indices(fd, num, &data.size);

	} else if( data.type == CHUNK_MATERIAL_DATA) {
	    chunk_pr(&data, "CHUNK_MATERIAL_DATA", level);

	    gs_read(fd, &num, sizeof(num), &data.size);
	    print(level+1, "NumberOfMaterials: %u\n", num);

	    while(data.size) {
		struct gs_chunk mat;
		gs_read_chunk(fd, &mat, &data.size);

		if(mat.type == CHUNK_MATERIAL_DEF)
		    dump_material(fd, &mat, level+1);
		else {
		    chunk_pr(&mat, 0, level+1);
		    lseek(fd, mat.size, SEEK_CUR);
		}
	    }
	} else {
	    chunk_pr(&data, 0, level);
	    lseek(fd, data.size, SEEK_CUR);
	    continue;
	}

	if(data.size) {
	    print(level, "THERE IS MORE DATA HERE\n");
	    lseek(fd, data.size, SEEK_CUR);
	}
    }
}

static void dump_gs_object(int fd, unsigned int size)
{
    struct gs_object info;
    struct gs_chunk obj;
    unsigned int i;

    gs_read(fd, &info, sizeof(info), &size);

    printf("  name: %.32s\n", info.name);
  /*  printf("  unknown data:");
    for(i = 0; i < 7; i++) {
	printf(" %.4g", info.unkn[i]);
	if(i == 2) printf("\n  unknown data:");
    } **/
    
    print(1, "pos:  [x: %.6g, y: %.6g, z: %.6g]\n",
        info.pos.x, info.pos.y, info.pos.z);
    print(1, "quat: [x: %.6g, y: %.6g, z: %.6g, w: %.6g]\n", 
	info.quat.x, info.quat.y, info.quat.z, info.quat.w);
    printf("\n");

    while(size) {
	gs_read_chunk(fd, &obj, &size);

	if(obj.type == CHUNK_OBJECT_DATA) {
	    chunk_pr(&obj, "CHUNK_OBJECT_DATA", 1);
	    dump_obj_data(fd, &obj.size, 2);
	} else if(obj.type == CHUNK_OBJ_PATH_INFO) {
	    dump_path_info(fd, &obj, 1);
	} else if(obj.type == CHUNK_OBJECT_DATA2) {
	    chunk_pr(&obj, "CHUNK_OBJ_MORE_VERTEX", 1);
	    dump_obj_data(fd, &obj.size, 2);
	}
	else {
	    chunk_pr(&obj, 0, 1);
	    lseek(fd, obj.size, SEEK_CUR);
	    continue;
	}

	if(obj.size) {
	    printf("  THERE IS MORE DATA HERE\n");
	    lseek(fd, obj.size, SEEK_CUR);
	}

    }
}


static void dump_gs_scene(int fd, unsigned int size)
{
    struct gs_scene_descr descr;
    struct gs_chunk scf;

    while(size) {
	gs_read_chunk(fd, &scf, &size);

	if(scf.type == CHUNK_SCENE_DESCR) {
	    chunk_pr(&scf, "CHUNK_SCENE_DESCR", 1);

	    gs_read(fd, &descr, sizeof(descr), &scf.size);
	    print(2, "unkn: %u name: '%.32s'\n", descr.unk, descr.name);
	    print(2, "file: '%.64s'\n", descr.file);
	    print(2, "offset: %.4g, %.4g, %.4g\n",
		   descr.ofX, descr.ofY, descr.ofZ);
	    print(2, "unkn: %#08x %#08x %#08x %#08x\n",
		   descr.u1, descr.u2, descr.u3, descr.u4);

	} else if(scf.type == CHUNK_SCENE_ORIG_FILE) {
	    chunk_pr(&scf, "CHUNK_SCENE_ORIG_FILE", 1);
	    char *str = malloc(scf.size);
	    gs_read(fd, str, scf.size, &scf.size);
	    print(2, "file: %s\n", str);
	    free(str);
	} else {
	    chunk_pr(&scf, 0, 1);
	    lseek(fd, scf.size, SEEK_CUR);
	    continue;
	}
	if(scf.size) {
	    print(1, "THERE IS MORE DATA HERE\n");
	    lseek(fd, scf.size, SEEK_CUR);
	}
    }
}


static void dump_gs_skeleton(int fd, unsigned int size)
{
    char name[32];
    uint32_t num_bones, i;
    struct gs_bone bone;

    gs_read(fd, name, sizeof(name), &size);
    gs_read(fd, &num_bones, sizeof(num_bones), &size);

    printf("  name: '%.32s'\n", name);
    printf("  NumberOfBones: %u\n", num_bones);

    while(size) {
	struct gs_chunk bch;
	gs_read_chunk(fd, &bch, &size);
	if(bch.type != CHUNK_BONE) {
	    chunk_pr(&bch, 0, 1);
	    lseek(fd, bch.size, SEEK_CUR);
	    continue;
	}

	chunk_pr(&bch, "CHUNK_BONE", 1);
	gs_read(fd, &bone, sizeof(bone), &bch.size);

	print(2, "name: %.32s n:%i", bone.name, bone.n);
	for(i = 0; i < 10; i++) {
	    if(i%4 == 0) printf("\n%*sunk:", 2*2, "");
	    printf(" %.4g", bone.u[i]);
	}
	printf("\n");
	while(bch.size) {
	    struct gs_chunk more;
	    gs_read_chunk(fd, &more, &bch.size);
	    if(more.type == CHUNK_BONE_OBJ_REF) {
		char obj_name[32];
		chunk_pr(&more, "CHUNK_BONE_OBJ_REF", 2);
		gs_read(fd, obj_name, sizeof(obj_name), &more.size);
		print(3, "name: %.32s\n", obj_name);

		if(more.size) {
		    print(3, "THERE IS MORE DATA HERE\n");
		    lseek(fd, more.size, SEEK_CUR);
		}
	    } else {
		chunk_pr(&more, 0, 2);
		lseek(fd, more.size, SEEK_CUR);
	    }
	}

    }
}

static void dump_gs_properties(int fd, unsigned int size)
{
    struct gs_object_properties prop;
    struct gs_chunk obj;
    
    gs_read(fd, &prop, sizeof(prop), &size);    
    
    printf("  name: %.32s\n", prop.name);
    printf("  className: %.32s\n", prop.className);
    printf("  sortAlpa: 0x%02x\n", prop.sortAlpa);
    printf("  defaultDrawDistance: 0x%02x\n", prop.defaultDrawDistance);
    printf("  minDrawDistance: %.6g\n", prop.minDrawDistance);
    printf("  maxDrawDistance: %.6g\n", prop.maxDrawDistance);

    while(size) {
	gs_read_chunk(fd, &obj, &size);

	if(obj.type == 0x22) {
	    chunk_pr(&obj, "CHUNK_OBJECT_DATA", 1);
	    dump_obj_data(fd, &obj.size, 2);
	}
	else {
	    chunk_pr(&obj, 0, 1);
	    lseek(fd, obj.size, SEEK_CUR);
	    continue;
	}

	if(obj.size) {
	    printf("  THERE IS MORE DATA HERE\n");
	    lseek(fd, obj.size, SEEK_CUR);
	}

    }
}


static void dump_gs_states_in(int fd, unsigned int *size, unsigned int level) {
    while(*size) {
	struct gs_chunk data;
	gs_read_chunk(fd, &data, size);

        if( data.type == CHUNK_GS_STATES_NAME) {
            chunk_pr(&data, "CHUNK_GS_STATES_NAME", level);
            char* buffer = malloc(data.size);
            gs_read(fd, buffer, data.size, &data.size);
            print(level, " name: %s\n", buffer);
            free(buffer);
            //lseek(fd, data.size, SEEK_CUR);
            continue;
	} else if( data.type == CHUNK_GS_STATES_ENTRY) {
            struct gs_states_entry state_entry;
            chunk_pr(&data, "CHUNK_GS_STATES_ENTRY", level);
            gs_read(fd, &state_entry, sizeof(state_entry), &data.size);
            print(level, " statename: %.32s\n", state_entry.name);
            dump_gs_states_in(fd, &data.size, level+1);
            //lseek(fd, data.size, SEEK_CUR);
            continue;
        } else if( data.type == CHUNK_GS_STATES_ENTRYSCRIPT) {   
            chunk_pr(&data, "CHUNK_GS_STATES_ENTRYSCRIPT", level);
            char* buffer = malloc(data.size);
            gs_read(fd, buffer, data.size, &data.size);
            print(level, " script: %s\n", buffer);
            free(buffer);
            continue;
        } else if( data.type == CHUNK_GS_STATES_ENTRYSCRIPT2) {   
            chunk_pr(&data, "CHUNK_GS_STATES_ENTRYSCRIPT2", level);
            char* buffer = malloc(data.size);
            gs_read(fd, buffer, data.size, &data.size);
            print(level, " script: %s\n", buffer);
            free(buffer);
            continue;
      	} else {
	    chunk_pr(&data, 0, level);
	    lseek(fd, data.size, SEEK_CUR);
	    continue;
	}

	if(data.size) {
	    print(level, "THERE IS MORE DATA HERE\n");
	    lseek(fd, data.size, SEEK_CUR);
	}
    }
}
static void dump_gs_states(int fd, unsigned int size)
{    
    struct gs_chunk states;
    
    while(size) {
	gs_read_chunk(fd, &states, &size);

	if(states.type == CHUNK_GS_STATES_NAMES ) {
	    chunk_pr(&states, "CHUNK_GS_STATES_NAMES", 2);
            dump_gs_states_in(fd, &states.size, 3);
        } else if(states.type == CHUNK_GS_STATES_SCRIPTS) {
	    chunk_pr(&states, "CHUNK_GS_STATES_SCRIPTS", 2);
            dump_gs_states_in(fd, &states.size, 3);
	} else {
	    chunk_pr(&states, 0, 3);
	    lseek(fd, states.size, SEEK_CUR);
	    continue;
	}
	if(states.size) {
	    print(2,"  THERE IS MORE DATA HERE\n");
	    lseek(fd, states.size, SEEK_CUR);
	}
    }
}




static void dump_gs_timeline_entry(int fd, struct gs_chunk *chunk, unsigned int level) {
    
    if(chunk->type == CHUNK_GS_TIMELINE_ENTRY) {
        struct gs_timeline_entry entry;
        chunk_pr_ext(chunk, "CHUNK_GS_TIMELINE_ENTRY", level, fd);
        gs_read(fd, &entry, sizeof(entry), &chunk->size);
        print(level," Days since 1920: %i (0x%08x)\n", entry.u1, entry.u1);
        print(level," Unknown 2:       0x%08x\n", entry.u2);
    } else if (chunk->type == CHUNK_GS_TIMELINE_EUNK1) {
        struct gs_timeline_eunk1 entryunk1;
        chunk_pr_ext(chunk, "CHUNK_GS_TIMELINE_EUNK1", level, fd);
        gs_read(fd, &entryunk1, sizeof(entryunk1), &chunk->size);
        print(level, " name: %.32s\n", entryunk1.objectname);
    } else if (chunk->type == CHUNK_GS_TIMELINE_EUNK2) {
        struct gs_timeline_eunk2 entryunk2;
        chunk_pr_ext(chunk, "CHUNK_GS_TIMELINE_EUNK2", level, fd);
        gs_read(fd, &entryunk2, sizeof(entryunk2), &chunk->size);
        print(level, " name: %.32s\n", entryunk2.objectname);
        print(level, " unkn: %#08x %#08x %#08x\n", entryunk2.u1,  entryunk2.u2, entryunk2.u3);
    } else if (chunk->type == CHUNK_GS_TIMELINE_EUNK4) {
        struct gs_timeline_eunk2 entryunk4;
        chunk_pr_ext(chunk, "CHUNK_GS_TIMELINE_EUNK4", level, fd);
        gs_read(fd, &entryunk4, sizeof(entryunk4), &chunk->size);
        print(level, " name: %.32s\n", entryunk4.objectname);
        print(level, " unkn: %#08x %#08x %#08x\n", entryunk4.u1,  entryunk4.u2, entryunk4.u3);
    } else if (chunk->type == CHUNK_GS_TIMELINE_EUNK8) {
        struct gs_timeline_eunk8 entryunk8;
        chunk_pr_ext(chunk, "CHUNK_GS_TIMELINE_EUNK8", level, fd);
        gs_read(fd, &entryunk8, sizeof(entryunk8), &chunk->size);
        print(level, " name: %.32s\n", entryunk8.objectname);
        print(level, " unkn: %#08x %#08x %#08x %#08x\n", entryunk8.u1,  entryunk8.u2, entryunk8.u3, entryunk8.u4);
    }

    while(chunk->size) {
	struct gs_chunk data;
	gs_read_chunk(fd, &data, &chunk->size);

        if( data.type == CHUNK_GS_TIMELINE_EUNK1 || 
            data.type == CHUNK_GS_TIMELINE_EUNK2 ||
            data.type == CHUNK_GS_TIMELINE_EUNK4 ||
            data.type == CHUNK_GS_TIMELINE_EUNK8) {
            dump_gs_timeline_entry(fd, &data, level+1);
            //chunk_pr(&data, "CHUNK_GS_TIMELINE_EUNK1", level);
            //lseek(fd, data.size, SEEK_CUR);
            continue;
	} else {
	    chunk_pr_ext(&data, 0, level+1, fd);
	    lseek(fd, data.size, SEEK_CUR);
	    continue;
	}

	if(data.size) {
	    print(level, "THERE IS MORE DATA HERE\n");
	    lseek(fd, data.size, SEEK_CUR);
	}
    }
}

static void dump_gs_timeline(int fd, unsigned int size)
{    
    struct gs_chunk tentry;
    struct gs_timeline timeline;

    gs_read(fd, &timeline, sizeof(timeline), &size);
    while(size) {
	gs_read_chunk(fd, &tentry, &size);

	if(tentry.type == CHUNK_GS_TIMELINE_ENTRY ) {
            dump_gs_timeline_entry(fd, &tentry, 4);
	} else {
	    chunk_pr(&tentry, 0, 3);
	    lseek(fd, tentry.size, SEEK_CUR);
	    continue;
	}
	if(tentry.size) {
	    print(2,"  THERE IS MORE DATA HERE\n");
	    lseek(fd, tentry.size, SEEK_CUR);
	}
    }
}


static void dump_gs_grid_objects(int fd, struct gs_chunk ch) {
    struct gs_grid_objects_header objects_header;
    struct gs_grid_objects_instance object_instance;
    unsigned long i = 0;
    unsigned long pos = 0;
    chunk_pr_ext(&ch, "CHUNK_GS_GRID_OBJECTS", 1, fd);
    gs_read(fd, &objects_header, sizeof (objects_header), &ch.size);
    print(2, " numobjectids:  %8i\n", objects_header.numobjectids);
    print(2, " numinstances:    %8i\n", objects_header.numinstances);


    while (ch.size) {
        struct gs_chunk childchunk;
        gs_read_chunk(fd, &childchunk, &ch.size);

        switch (childchunk.type) {
            case CHUNK_GS_GRID_LISTOBJECTS:
                chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_LISTOBJECTS", 2, fd);
                char *buffer = calloc(childchunk.size + 10, 1);
                if (!buffer) die("Can't alloc memory");
                if (childchunk.size > 0) {
                    gs_read(fd, buffer, childchunk.size, &childchunk.size);
                    int count = 0;
                    while (*buffer) {
                        print(3, "%4i: %s\n", count, buffer);
                        while (*(buffer++)); // eat all chars and zero
                        count += 1;
                    }
                    if (count != objects_header.numobjectids) {
                        print(3, "*warn* count mismatch, %u should be %u", count, objects_header.numobjectids);
                    }
                    free(buffer);
                }
                break;
            case CHUNK_GS_GRID_OBJ_INSTANCE:
                chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_OBJ_INSTANCE", 2, fd);          
                if (verbose == 0) {
                       print(3, "<...use -l to see more data here>\n");
                }
                pos = tell(fd);
                for (i = 0; i < objects_header.numinstances; i++) {
                    gs_read(fd, &object_instance, sizeof (object_instance), &childchunk.size);
                    if (verbose == 1) {
                        print(3, "%4i:\n", i);
                        if (fileoffsets) {
                            print(3, " @ %#0.8x\n", pos);
                            pos += sizeof(object_instance);
                        }
                        print(3, "     ref: %hu\n", (object_instance.refnum & 0x3FFF));
                        print(3, "       x: %08hi y:   %08hi z: %08hi\n", object_instance.posX, object_instance.posY, object_instance.posZ);
                        print(3, "      u1: 0x%08x u2: 0x%08x\n", object_instance.u1, object_instance.u2);
                    }
                    if ((unsigned long) (object_instance.refnum & 0x3FFF) >= objects_header.numobjectids) {
                        print(3, "*warn* reference number for entry %i broken\n", i);
                    }
                }
                break;
            case CHUNK_GS_GRID_OBJ_INSTANCE_NAMES:
                chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_OBJ_INSTANCE_NAMES", 2, fd);
                lseek(fd, childchunk.size, SEEK_CUR);
                break;
            default:
                chunk_pr_ext(&childchunk, 0, 2, fd);
                print(3, "*warn* this chunk shouldn't be here\n");
                lseek(fd, childchunk.size, SEEK_CUR);
        }

    }
}

static void dump_gs_grid(int fd, unsigned int size)
{
    int verboselist = 0;
    int listcounter_a;
    
    struct gs_grid grid;
    struct gs_grid_objects_header objects_header;
    struct gs_grid_commonlist_header commonlist_header;
    struct gs_grid_unknown6_header unknownheader6;
    
    
    gs_read(fd, &grid, sizeof(grid), &size);
    printf(" Unknown 1 (Map Size?):       0x%08x\n", grid.u1);
    printf(" Unknown 2 (Map Size?):       0x%08x\n", grid.u2);
    printf(" Unknown 3:                   0x%08x\n", grid.u3);
    printf(" Unknown 4:                   0x%08x\n", grid.u4);
    printf(" Unknown 5:                   0x%08x\n", grid.u5);
    printf(" Unknown 6:                   0x%08x\n", grid.u6);
    
    while(size) {
	struct gs_chunk ch;
        
	gs_read_chunk(fd, &ch, &size);
        
        switch (ch.type) {
            case CHUNK_GS_GRID_MAP:
                chunk_pr_ext(&ch, "CHUNK_GS_GRID_MAP", 1, fd);
                lseek(fd, ch.size, SEEK_CUR);
                break;
            case CHUNK_GS_GRID_OBJECTS:
                dump_gs_grid_objects(fd, ch);
                break;
            case CHUNK_GS_GRID_UNKNOWN4:
                verboselist = 0;
                listcounter_a = 0;
                chunk_pr_ext(&ch, "CHUNK_GS_GRID_UNKNOWN4", 1, fd);
                gs_read(fd, &commonlist_header, sizeof(commonlist_header), &ch.size);
                print(2, " numvehiclenames    %8i\n", commonlist_header.numlist);
                print(2, " Unknown 2:       0x%0.8x\n", commonlist_header.u2);
           
                while(ch.size) {
                    struct gs_chunk childchunk;
                    gs_read_chunk(fd, &childchunk, &ch.size);
                    
                    switch (childchunk.type) {
                        case CHUNK_GS_GRID_LISTVEHICLES:
                            chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_LISTVEHICLES", 2, fd);
                            char *buffer = calloc(childchunk.size+10, 1);
                            if (!buffer) die("Can't alloc memory");
                            if (childchunk.size > 0) {
                               gs_read(fd, buffer, childchunk.size, &childchunk.size);
                               int count = 0;
                               while(*buffer) {
                                   count+=1;
                                   print(3, " %8i: %s\n", count, buffer);
                                   while(*(buffer++));  // eat all chars and zero
                               }
                               if (count != commonlist_header.numlist) {
                                   print(3, "*warn* count mismatch, %u should be %u", count, commonlist_header.numlist);
                               }
                               free(buffer);
                            }
                            break;
                        case CHUNK_GS_GRID_UNKNOWN4_2:
                            listcounter_a++;
                            // fall through
                        default:
                            if (verbose == 0) {
                                if(verboselist == 0) {
                                    chunk_pr_ext(&childchunk, 0, 2, fd);
                                    print(3, "<...use -l to see more data here>\n");
                                    verboselist = 1;
                                }
                            } else {
                                 chunk_pr_ext(&childchunk, 0, 2, fd);
                            }
                            lseek(fd, childchunk.size, SEEK_CUR);
                    }
                }
                if (listcounter_a != commonlist_header.u2) {
                        print(3, "*warn* CHUNK(0x07040200) count should be %u but found %u.\n\n", commonlist_header.u2, listcounter_a);
                }
                break;
// ---------------------
           case CHUNK_GS_GRID_UNKNOWN5:
                verboselist = 0;
                listcounter_a = 0;
                chunk_pr_ext(&ch, "CHUNK_GS_GRID_UNKNOWN5", 1, fd);
                gs_read(fd, &commonlist_header, sizeof(commonlist_header), &ch.size);
                print(2, " numcitizennames:   %8i\n", commonlist_header.numlist);
                print(2, " Unknown 2:       0x%0.8x\n", commonlist_header.u2);
           
                while(ch.size) {
                    struct gs_chunk childchunk;
                    gs_read_chunk(fd, &childchunk, &ch.size);
                    
                    switch (childchunk.type) {
                        case CHUNK_GS_GRID_LISTCITIZEN:
                            chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_LISTCITIZEN", 2, fd);
                            char *buffer = calloc(childchunk.size+10, 1);
                            if (!buffer) die("Can't alloc memory");
                            if (childchunk.size > 0) {
                               gs_read(fd, buffer, childchunk.size, &childchunk.size);
                               int count = 0;
                               while(*buffer) {
                                   count+=1;
                                   print(3, " %8i: %s\n", count, buffer);
                                   while(*(buffer++));  // eat all chars and zero
                               }
                               if (count != commonlist_header.numlist) {
                                   print(3, "*warn* count mismatch, %u should be %u", count, commonlist_header.numlist);
                               }
                               free(buffer);
                            }
                            break;
                        case CHUNK_GS_GRID_UNKNOWN5_2:
                            listcounter_a++;
                            // fall through
                        default:
                            if (verbose == 0) {
                                if(verboselist == 0) {
                                    chunk_pr_ext(&childchunk, 0, 2, fd);
                                    print(3, "<...use -l to see more data here>\n");
                                    verboselist = 1;
                                }
                            } else {
                                 chunk_pr_ext(&childchunk, 0, 2, fd);
                            }
                            lseek(fd, childchunk.size, SEEK_CUR);
                    }
                }
                if (listcounter_a != commonlist_header.u2) {
                       print(3, "*warn* CHUNK(0x07040500) count should be %u but found %u.\n\n", commonlist_header.u2, listcounter_a);
                }
                break;
// ---------------------
           case CHUNK_GS_GRID_UNKNOWN6:
                listcounter_a = 0;
                verboselist = 0;
                chunk_pr_ext(&ch, "CHUNK_GS_GRID_UNKNOWN6", 1, fd);
                gs_read(fd, &unknownheader6, sizeof(unknownheader6), &ch.size);
                print(2, " numplayervehicles:%8i\n", unknownheader6.numplayervehiclenames);
                print(2, " Unknown 2:       0x%0.8x\n", unknownheader6.u2);
                print(2, " Unknown 3:       0x%0.8x\n", unknownheader6.u3);
                print(2, " Unknown 4:       0x%0.8x\n", unknownheader6.u4);
                
                while(ch.size) {
                    struct gs_chunk childchunk;
                    gs_read_chunk(fd, &childchunk, &ch.size);
                    
                    switch (childchunk.type) {
                        case CHUNK_GS_GRID_LISTPLAYERVEHICLES:
                            chunk_pr_ext(&childchunk, "CHUNK_GS_GRID_LISTPLAYERVEHICLES", 2, fd);
                            char *buffer = calloc(childchunk.size+10, 1);
                            if (!buffer) die("Can't alloc memory");
                            if (childchunk.size > 0) {
                               gs_read(fd, buffer, childchunk.size, &childchunk.size);
                               int count = 0;
                               while(*buffer) {
                                   count+=1;
                                   print(3, " %8i: %s\n", count, buffer);
                                   while(*(buffer++));  // eat all chars and zero
                               }
                               if (count != unknownheader6.numplayervehiclenames) {
                                   print(3, "*warn* count mismatch, %u should be %u", count, unknownheader6.numplayervehiclenames);
                               }
                               free(buffer);
                            }
                            break;
                        case 0x7060200:
                            listcounter_a++;
                        default:
                            if (verbose == 0) {
                                if(verboselist == 0) {
                                    chunk_pr_ext(&childchunk, 0, 2, fd);
                                    print(3, "<...use -l to see more data here>\n");
                                    verboselist = 1;
                                }
                            } else {
                                 chunk_pr_ext(&childchunk, 0, 2, fd);
                            }
                            lseek(fd, childchunk.size, SEEK_CUR);
                    }
                }
                if (listcounter_a != unknownheader6.u2) {
                       print(3, "*warn* CHUNK(0x07060200) count should be %u but found %u.\n\n", commonlist_header.u2, listcounter_a);
                }                
                break;
          case CHUNK_GS_GRID_UNKNOWN7:
                listcounter_a = 0;
                verboselist = 0;
                chunk_pr_ext(&ch, "CHUNK_GS_GRID_UNKNOWN7", 1, fd);
                // skip 0x90 bytes
                lseek(fd, 0x90, SEEK_CUR);
                ch.size -= 0x90;
  
                while(ch.size) {
                    struct gs_chunk childchunk;
                    gs_read_chunk(fd, &childchunk, &ch.size);
                    
                    switch (childchunk.type) {
                        default:
                            if (verbose < 2) {
                                if(verboselist == 0) {
                                    chunk_pr_ext(&childchunk, 0, 2, fd);
                                    print(3, "<...use -l2 to see more data here>\n");
                                    verboselist = 1;
                                }
                            } else {
                                 chunk_pr_ext(&childchunk, 0, 2, fd);
                            }
                            lseek(fd, childchunk.size, SEEK_CUR);
                    }
                }       
                break;
            default:
                chunk_pr_ext(&ch, 0, 1, fd);
                lseek(fd, ch.size, SEEK_CUR);
        }

    }

}


void usage(char* name) {
    printf("Usage: %s <options> <file>\n"
    " -l  verbose output\n"
    " -l2 even more output\n"
    " -o  output fileoffsets\n", name);
    
}

int main(int argc, char **argv)
{
    unsigned int i;
    char *filename = NULL;

    printf(TOOLNAME "\n");
    
    for(i=1; i<argc; i++) {
        char *arg = argv[i];
        if(arg[0] == '-') {
            if (arg[1] == 'l') {
                verbose = 1;
                if (arg[2] == '2') {
                    verbose = 2;
                }
            } else if (arg[1] == 'o') {
                fileoffsets = 1;
            } else {
                printf("Unknown option %s \n", argv[i]);
                usage(argv[0]);
                return -1;
            }
	} else {
            filename = argv[i];
        }
    }
    
    
    if(!filename) {
        usage(argv[0]);
	return -1;
    }

    int fd = open(filename, OPEN_FLAGS);
    if(fd == -1)
	die("open(): %s\n", strerror(errno));

    struct gs_chunk file;
    read(fd, &file, sizeof(file));
    switch(file.type) {
    case FILETYPE_GS_ARCHIVE:
        printf("File Type: GS_ARCHIVE(%#08x) (%u bytes)\n",
	       file.type, file.size);
	dump_gs_archive(fd, file.size);
	break;
    case FILETYPE_GS_OBJECT:
	printf("File Type: GS_OBJECT(%#08x) (%u bytes)\n",
	       file.type, file.size);
	dump_gs_object(fd, file.size);
	break;
    case FILETYPE_GS_MATERIAL:
	printf("File Type: GS_MATERIAL(%#08x) (%u bytes)\n",
	       file.type, file.size);
	dump_material(fd, &file, 1);
	break;
     case FILETYPE_GS_SCENE:
	printf("File Type: GS_SCENE(%#08x) (%u bytes)\n",
	       file.type, file.size);
	dump_gs_scene(fd, file.size);
	break;
    case FILETYPE_GS_SKELETON:
	printf("File Type: GS_SKELETON(%#08x) (%u bytes)\n",
	       file.type, file.size);
	dump_gs_skeleton(fd, file.size);
	break;
    case FILETYPE_GS_ANIMATION:
	printf("File Type: GS_ANIMATION(%#08x) (%u bytes)\n",
	       file.type, file.size);
	break;
    case FILETYPE_GS_PROPERTIES:
	printf("File Type: GS_PROPERTIES(%#08x) (%u bytes)\n",
	       file.type, file.size);
        dump_gs_properties(fd, file.size);
	break;
    case FILETYPE_GS_GRID:
	printf("File Type: GS_GRID(%#08x) (%u bytes)\n",
	       file.type, file.size);
        dump_gs_grid(fd, file.size);
	break;
    case FILETYPE_GS_TIMELINE:
	printf("File Type: GS_TIMELINE(%#08x) (%u bytes)\n",
	       file.type, file.size);
        dump_gs_timeline(fd, file.size);
	break;
    case FILETYPE_GS_STATES:
	printf("File Type: GS_STATES(%#08x) (%u bytes)\n",
	       file.type, file.size);
        dump_gs_states(fd, file.size);
	break;
    default:
	printf("File type: %#08x (%u)\n", file.type, file.size);
    }

    return 0;
}
